/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *      cmds.c
 *
 *      This rather bulky module contains all the functions that implement
 *      commands.  It also handles initialization of the interface to those
 *      functions.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/Xaw/Label.h>

#include "calc.h"
#include "types.h"
#include "../siod/siod.h"
#include "../xstuff/fonts.h"

typedef struct {
	char *name;
	 LISP(*function) ();
} _fn_table;

extern _fn_table fn_table[];

/* moving around */

static char *testlist[] = {	"testitem0",
	"testitem1", "testitem2", "testitem3", "testitem4", "testitem5",
	"testitem6", "testitem7", "testitem8", "testitem9"
};

static LISP listsel_test()
{
	int i = listsel(app_context, topLevel, "Choose an item", testlist, 10);
	if (i < 0)
		printf("User clicked Cancel\n");
	else
		printf("User chose %d\n", i);
	return NIL;
}

static LISP keyboard_quit()
{
	llpr("Quit");
	return NIL;
}

static LISP what_cursor_position()
{
	char b[256], *type;
	spread **matrix = w_list->buf->matrix;
	int row = w_list->point_pos.row, col = w_list->point_pos.col;

	switch (ret_type(matrix, row, col)) {
	case EMPTY:
		type = "EMPTY";
		break;
	case CONSTANT:
		type = "CONSTANT";
		break;
	case LABEL:
		type = "LABEL";
		break;
	case EXPRESSION:
		type = "EXPRESSION";
		break;
	default:
		type = "ERROR";
	}
	sprintf(b, "(%d, %d) is %s: %s", row, col, type,
		ret_type(matrix, row, col) == EMPTY ? "(null)"
		: ret_taddr(matrix, row, col));
	llpr(b);
	return NIL;
}

static LISP edit_label()
{
	char b[256];
	int i;
	spread **matrix = w_list->buf->matrix;
	int row = w_list->point_pos.row, col = w_list->point_pos.col;

	strcpy(b, ret_type(matrix, row, col) == EMPTY ? ""
	       : ret_taddr(matrix, row, col));
	if (ask_for_str("Label: ", b)) {
		if ((i = pool_install(b))) {
			if (ins_data(matrix, i, (double) 0, LABEL, row, col))
				w_list->buf->change = TRUE;
			pr_scr_flag = TRUE;
		} else
			llpr("String pool full");
	}
	return NIL;
}

static int recalc = 1;

static LISP auto_recalc(LISP n)
{
	if (NFLONUMP(n))
		err("wta(1st) to auto-recalc", n);
	recalc = FLONM(n);
	return NIL;
}

static LISP recalc_matrix()
{
	llpr("calculating...");
	calc_matrix(w_list->buf->matrix);
	llpr("calculating... done");
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP edit_expression()
{
	char b[256];
	int i;
	double value;
	spread **matrix = w_list->buf->matrix;
	int row = w_list->point_pos.row, col = w_list->point_pos.col;

	strcpy(b, ret_type(matrix, row, col) == EMPTY ? ""
	       : ret_taddr(matrix, row, col));
	if (ask_for_str("Expression: ", b)) {
		errorflag = FALSE;
		value = parse_expr(matrix, b, row, col);
		if (errorflag) {
#if 0	/* this overwrites the message from siod, which is more useful */
			llpr("Syntax error in expression");
#endif
			errorflag = FALSE;
		} else {
			if ((i = pool_install(b))) {
				if (ins_data(matrix, i, value,
				  is_constant(b) ? CONSTANT : EXPRESSION,
					     row, col)) {
					int i;

					w_list->buf->change = TRUE;
					for (i = 0; i < recalc; i++)
						(void)recalc_matrix();
				}
				pr_scr_flag = TRUE;
			} else
				llpr("String pool full");
		}
	}
	return NIL;
}

static LISP recenter()
{
	/* we don't want to do this anymore, just redraw the screen */
	pr_scr_flag = TRUE;
	return NIL;
}

/* move point inside visible area */
static void fix_point(int width, int height)
{
	int cur_x, cur_y;

	get_cell_coords(w_list, w_list->top.row, w_list->top.col,
			w_list->point_pos.row, w_list->point_pos.col,
			&cur_x, &cur_y);
	while (cur_y < 0) {
		w_list->point_pos.row++;
		cur_y += 20;
	}
	while (cur_y > height - 20) {
		w_list->point_pos.row--;
		cur_y -= 20;
	}
	while (cur_x < 0) {
		w_list->point_pos.col++;
		cur_x += 80;
	}
	while (cur_x > width - 80) {
		w_list->point_pos.col--;
		cur_x -= 80;
	}
}

static LISP scroll_up()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = -height + 20;
	new_row = w_list->top.row;
	while (new_y < 0) {
		new_y += 20;
		new_row--;
	}
	if (new_row < 1)
		new_row = 1;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_cell_up()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = -20;
	new_row = w_list->top.row;
	while (new_y < 0) {
		new_y += 20;
		new_row--;
	}
	if (new_row < 1)
		new_row = 1;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_down()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = height - 20;
	new_row = w_list->top.row;
	while (new_y > 0) {
		new_y -= 20;
		new_row++;
	}
	if (new_row > max_lines)
		new_row = max_lines;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_cell_down()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_y, new_row;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_y = 20;
	new_row = w_list->top.row;
	while (new_y > 0) {
		new_y -= 20;
		new_row++;
	}
	if (new_row > max_lines)
		new_row = 1;
	w_list->top.row = new_row;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_left()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_x, new_col;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_x = -width + 80;
	new_col = w_list->top.col;
	while (new_x < 0) {
		new_x += 80;
		new_col--;
	}
	if (new_col < 1)
		new_col = 1;
	w_list->top.col = new_col;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_cell_left()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_x, new_col;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_x = -80;
	new_col = w_list->top.col;
	while (new_x < 0) {
		new_x += 80;
		new_col--;
	}
	if (new_col < 1)
		new_col = 1;
	w_list->top.col = new_col;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_right()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_x, new_col;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_x = width - 80;
	new_col = w_list->top.col;
	while (new_x > 0) {
		new_x -= 80;
		new_col++;
	}
	if (new_col > max_columns)
		new_col = max_columns;
	w_list->point_pos.col = w_list->top.col = new_col;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP scroll_cell_right()
{
	Window root, cell_win = XtWindow(w_list->grid);
	int x, y, new_x, new_col;
	unsigned int width, height, border_width, depth;

	XGetGeometry(display, cell_win, &root, &x, &y, &width, &height,
		     &border_width, &depth);
	new_x = 80;
	new_col = w_list->top.col;
	while (new_x > 0) {
		new_x -= 80;
		new_col++;
	}
	if (new_col > max_columns)
		new_col = max_columns;
	w_list->top.col = new_col;
	fix_point(width, height);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP delete_cell()
{
	if (ins_data(w_list->buf->matrix, 0, (double) 0, EMPTY,
		     point_position(w_list).row,
		     point_position(w_list).col))
		w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP insert_line()
{
	downshift_matrix(w_list->buf, point_position(w_list).row);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP remove_line()
{
	upshift_matrix(w_list->buf, point_position(w_list).row);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP insert_col()
{
	rightshift_matrix(w_list->buf, point_position(w_list).col);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP remove_col()
{
	leftshift_matrix(w_list->buf, point_position(w_list).col);
	w_list->buf->change = TRUE;
	pr_scr_flag = TRUE;
	return NIL;
}

/* Start a new Siag */
static LISP new_siag()
{
	if (!fork()) {
		/* this is the child's fork */
		execlp("siag", "Siag", NULL);
		/* shouldn't return, but if it does we don't want to keep it */
		exit(0);
	}
	return NIL;
}

static void do_help(helpfile)
char *helpfile;
{
	char b[256];
	sprintf(b, "file://localhost%s/%s", siaghome, helpfile);
	if (!fork()) {
		execlp(siaghelp, "Siaghelp", "-geometry", "523x356", b, NULL);
		exit(0);
	}
}

static LISP help_contents()
{
	do_help("siag/docs/siag.html");
	return NIL;
}

static LISP help_copyright()
{
	do_help("siag/docs/Copying");
	return NIL;
}


static LISP help_for_help()
{
	do_help("siag/docs/siaghelp.html");
	return NIL;
}

/* Windows and buffers */

static char *format[] = {
	"Siag (*.siag)",
	"All files (*)",
	NULL};

/*X
   static void load_buffer()

   Load a buffer from file.
   X */
static LISP load_buffer()
{
	char path[FILENAME_MAX], name[FILENAME_MAX];
	char fn[FILENAME_MAX];
	spread **matrix;
	buffer *b;

	getcwd(path, FILENAME_MAX);
	strcpy(name, "");
	fn[0] = '\0';
#if 0	/* plain dialog box */
	if (ask_for_str("File name: ", fn)) {
#else	/* file selector box */
	if (fsel_input(topLevel, path, name, format)) {
		sprintf(fn, "%s/%s", path, name);
#endif
		matrix = new_matrix();
		b = new_buffer(matrix, buffer_name(fn), fn);
		llpr("Loading");

		if (loadmatrix(fn, b))
			llpr("New file");
	/* this will fail if load returns 1 because the string pool is full */

		calc_matrix(matrix);
		b->change = FALSE;
		w_list->buf = b;
		pr_scr_flag = TRUE;
	}
	return NIL;
}

static LISP
delete_window()
{
	if (!remove_window(w_list))
		llpr("Attempt to delete sole ordinary window");
	else
		pr_scr_flag = TRUE;
	return NIL;
}

static LISP
delete_other_windows()
{
	while (remove_window(w_list->next));
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP
split_window_vertically()
{
	if (!split_window(w_list))
		llpr("This window is too small to split");
	else
		pr_scr_flag = TRUE;
	return NIL;
}

static LISP
other_window()
{
	w_list = w_list->next;
	pr_scr_flag = TRUE;
	return NIL;
}

/*X
   static void save_buffer()

   Save the buffer in the currently active window to file.
   X */
static LISP save_buffer()
{
	llpr("Saving");

	if (savematrix(w_list->buf->path, w_list->buf))
		llpr("Couldn't save");
	else {
		w_list->buf->change = FALSE;
		pr_scr_flag = TRUE;
	}
	return NIL;
}

/*X
   static void save_buffer_as()

   Save the buffer in the currently active window to a named file.
   X */
static LISP save_buffer_as()
{
	char path[FILENAME_MAX], name[FILENAME_MAX];
	char fn[FILENAME_MAX];

#if 0
	strcpy(fn, w_list->buf->path);
#endif
	getcwd(path, FILENAME_MAX);
	strcpy(name, "");
	fn[0] = '\0';
#if 0	/* plain dialog box */
	if (ask_for_str("Save as: ", fn)) {
#else	/* file selector box */
	if (fsel_input(topLevel, path, name, format)) {
		sprintf(fn, "%s/%s", path, name);
#endif
		llpr("Saving");
		if (savematrix(fn, w_list->buf))
			llpr("Couldn't save");
		else {
			w_list->buf->change = FALSE;
			strcpy(w_list->buf->path, fn);
		}
		pr_scr_flag = TRUE;
	}
	return NIL;
}

#if USE_COMPLETION
/*X
   static void complete_name(char *name)
   This function takes a partially completed buffer name
   and returns the first buffer name that matches it.
   X */
static int complete_name(name)
char *name;
{
	buffer *b;
	int len;

	b = w_list->buf;	/* start with the next buffer */
	do {
		b = b->next;
		if ((len = strlen(name)) == 0 
				|| !strncmp(b->name, name, len)) {
			strcpy(name, b->name);
			return FALSE;
		}
	} while (b != w_list->buf);
	return FALSE;
}
#endif

static LISP
switch_to_buffer()
{
	buffer *b;
	char *blist[100];
	int nblist = 0, n;

	b = w_list->buf;
	do {
		b = b->next;
		blist[nblist++] = b->name;
	} while (b != w_list->buf);
	if ((n = listsel(app_context, topLevel, "Change Buffer:", blist, nblist)) >= 0)
		w_list->buf = find_buffer_by_name(blist[n]);
	return NIL;
}

static LISP
kill_buffer()
{
	buffer *b, *next_b;
	window *w;
	char *blist[100];
	int nblist = 0, n;

	b = w_list->buf;
	do {
		b = b->next;
		blist[nblist++] = b->name;
	} while (b != w_list->buf);
	if ((n = listsel(app_context, topLevel, "Kill Buffer:", blist, nblist)) >= 0) {
		if ((b = find_buffer_by_name(blist[n])) != NULL) {
			if (b != b->next) {
				next_b = free_buffer(b);
				w = w_list;
				do {
					if (w->buf == b)
						w->buf = next_b;
					w = w->next;
				} while (w != NULL && w != w_list);
				pr_scr_flag = TRUE;
			}
			else llpr("Couldn't kill last buffer");
		}
	}
	return NIL;
}

static void
copy_from_start(spread **matrix, position blku, position blkl, int nr, int nk)
{
	int i, j;

	for (i = 0; i < nr; i++) {
		for (j = 0; j < nk; j++) {
			if (ins_data(matrix,
				     ret_text(matrix, blku.row + i, blku.col + j),
				     ret_val(matrix, blku.row + i, blku.col + j),
				     ret_type(matrix, blku.row + i, blku.col + j),
				     w_list->point_pos.row + i,
				     w_list->point_pos.col + j)) {
				w_list->buf->change = TRUE;
			}
		}
	}
}

static void
copy_from_end(spread **matrix, position blku, position blkl, int nr, int nk)
{
	int i, j;

	for (i = nr - 1; i >= 0; i--) {
		for (j = nk - 1; j >= 0; j--) {
			if (ins_data(matrix,
				     ret_text(matrix, blku.row + i, blku.col + j),
				     ret_val(matrix, blku.row + i, blku.col + j),
				     ret_type(matrix, blku.row + i, blku.col + j),
				     w_list->point_pos.row + i,
				     w_list->point_pos.col + j)) {
				w_list->buf->change = TRUE;
			}
		}
	}
}

static LISP 
copy_block()
{
	int nr, nk;
	spread **matrix = w_list->buf->matrix;
	position blku, blkl;

	blku = w_list->blku;
	blkl = w_list->blkl;
	nr = blkl.row - blku.row + 1;
	nk = blkl.col - blku.col + 1;
	copy_from_start(matrix, blku, blkl, nr, nk);
	calc_matrix(matrix);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP
intelligent_copy_block()
{
	int nr, nk;
	spread **matrix = w_list->buf->matrix;
	position blku, blkl;

	blku = w_list->blku;
	blkl = w_list->blkl;
	nr = blkl.row - blku.row + 1;
	nk = blkl.col - blku.col + 1;
	if (w_list->point_pos.row < blku.row ||
	    (w_list->point_pos.row == blku.row &&
	     w_list->point_pos.col < blku.col)) {
		copy_from_start(matrix, blku, blkl, nr, nk);
	} else {
		copy_from_end(matrix, blku, blkl, nr, nk);
	}
	calc_matrix(matrix);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP delete_block()
{
	register int i, j;
	spread **matrix = w_list->buf->matrix;
	position blku, blkl;

	blku = w_list->blku;
	blkl = w_list->blkl;
	for (i = blku.row; i <= blkl.row; i++)
		for (j = blku.col; j <= blkl.col; j++)
			if (ins_data(matrix, 0, (double) 0, EMPTY, i, j))
				w_list->buf->change = TRUE;
	calc_matrix(matrix);
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP plot_block(LISP style)
{
	int sty;

	if (NFLONUMP(style))
		err("wta(1st) to plot-block", style);
	sty = FLONM(style);
	plot(sty);
	return NIL;
}

static LISP preview_block()
{
	preview();
	return NIL;
}

static LISP print_block()
{
	int nr, nk;

	nr = w_list->blkl.row - w_list->blku.row + 1;
	nk = w_list->blkl.col - w_list->blku.col + 1;
	printer((char *) 0, w_list->buf->matrix,
		w_list->blku.row, w_list->blku.col, nr, nk);
	return NIL;
}

static LISP set_block()
{
	if (point_position(w_list).row < mark_position(w_list).row) {
		w_list->blku.row = point_position(w_list).row;
		w_list->blkl.row = mark_position(w_list).row;
	} else {
		w_list->blku.row = mark_position(w_list).row;
		w_list->blkl.row = point_position(w_list).row;
	}
	if (point_position(w_list).col < mark_position(w_list).col) {
		w_list->blku.col = point_position(w_list).col;
		w_list->blkl.col = mark_position(w_list).col;
	} else {
		w_list->blku.col = mark_position(w_list).col;
		w_list->blkl.col = point_position(w_list).col;
	}
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP block_borders(LISP style)
{
	int sty, r, c, fmt;
	spread **matrix = w_list->buf->matrix;

	if (NFLONUMP(style))
		err("wta(1st) to block-borders", style);
	sty = FLONM(style);
	for (r = w_list->blku.row; r <= w_list->blkl.row; r++)
		for (c = w_list->blku.col; c <= w_list->blkl.col; c++) {
			fmt = ret_format(matrix, r, c);
			switch (sty) {
			case 1:		/* borders */
				if (r == w_list->blku.row)
					fmt |= BORDER_TOP;
				if (r == w_list->blkl.row)
					fmt |= BORDER_BOTTOM;
				if (c == w_list->blku.col)
					fmt |= BORDER_LEFT;
				if (c == w_list->blkl.col)
					fmt |= BORDER_RIGHT;
				break;
			case 2:		/* grid */
				fmt |= BORDER_MASK;
				break;
			default:	/* none */
				fmt &= ~BORDER_MASK;
			}
			ins_format(matrix, r, c, fmt);
		}
	pr_scr_flag = TRUE;
	return NIL;
}

static LISP save_block_file()
{
	char fn[256];
	int nr, nk;

	fn[0] = '\0';
	if (ask_for_str("File name: ", fn)) {
		nr = w_list->blkl.row - w_list->blku.row + 1;
		nk = w_list->blkl.col - w_list->blku.col + 1;
		printer(fn, w_list->buf->matrix,
			w_list->blku.row, w_list->blku.col, nr, nk);
	}
	return NIL;
}

/* This actually uses arguments, so will be difficult to lispify */
#if 0
static void print(argc, argv)
int argc;
char *argv[];
{
	char b[256], i;

	*b = 0;
	for (i = 0; i < argc; i++) {
		strcat(b, argv[i]);
		strcat(b, " ");
	}
	llpr(b);
}
#endif

static LISP execute_extended_command()
{
	char b[256];

	b[0] = '\0';
	if (ask_for_str("Cmd: ", b))
		execute(b);
	return NIL;
}

static char *quit_buttons[] = {"Yes", "No", "Cancel"};

static LISP quit_siag()
{
	char prompt[256];
	buffer *b = b_list;
	do {
		if (b->change) {
			sprintf(prompt, "Save %s?", b->name);
			switch (alertbox(app_context, topLevel,
					prompt, quit_buttons, 3)) {
			case 0:
				savematrix(b->path, b);
				break;
			case 2:
				return NIL;
			default:
				break;
			}
		}
		b = b->next;
	} while (b != b_list);
	exit(0);
	return NIL;
}

static LISP go_to()
{
	char b[256];
	int tr = 0, tc = 0;

	b[0] = '\0';
	if (ask_for_str("Go to: ", b))
		sscanf(b, "%d %d", &tr, &tc);

	if ((tr > 0) && (tr <= max_lines))
		w_list->point_pos.row = tr;
	if ((tc > 0) && (tc <= max_columns))
		w_list->point_pos.col = tc;
	return NIL;
}

static LISP set_cell_width()
{
	char b[256];
	int width = 0;

	b[0] = 0;
	if (ask_for_str("Width: ", b))
		sscanf(b, "%d", &width);

	if (width > 10 && width < 150) {
		w_list->buf->width[w_list->point_pos.col] = width;
		pr_scr_flag = 1;
	}
	return NIL;
}

static LISP set_cell_height()
{
	char b[256];
	int height = 0;

	b[0] = 0;
	if (ask_for_str("Height: ", b))
		sscanf(b, "%d", &height);

	if (height > 5 && height < 100) {
		w_list->buf->height[w_list->point_pos.row] = height;
		pr_scr_flag = 1;
	}
	return NIL;
}

static LISP set_block_format()
{
	int format = 0, mask = 0;

	format = ret_format(w_list->buf->matrix,
		w_list->blku.row, w_list->blku.col);
	if (font_input(&format, &mask)) {
		int r, c;

		for (r = w_list->blku.row; r <= w_list->blkl.row; r++)
			for (c = w_list->blku.col; c <= w_list->blkl.col; c++)
				ins_format(w_list->buf->matrix,
					r, c, format);
		pr_scr_flag = 1;
	}
	return NIL;
}

static LISP set_cell_format()
{
	int format = 0, mask = 0;

	format = ret_format(w_list->buf->matrix, w_list->point_pos.row,
			    w_list->point_pos.col);
	if (font_input(&format, &mask)) {
		ins_format(w_list->buf->matrix, w_list->point_pos.row,
			   w_list->point_pos.col, format);
		pr_scr_flag = 1;
	}
	return NIL;
}

static LISP no_op()
{
	llpr("This command does nothing");
	return NIL;
}

static LISP print_version()
{
	llpr(version);
	return NIL;
}

/* These functions allow implementation of commands in Scheme
	rather than in C with Scheme wrappers */
static LISP get_point_row()
{
	return flocons(w_list->point_pos.row);
}

static LISP get_point_col()
{
	return flocons(w_list->point_pos.col);
}

static LISP set_point_row(LISP row)
{
	int n;

	if (NFLONUMP(row))
		err("wta(1st) to set-point-row", row);
	n = FLONM(row);
	w_list->point_pos.row = n;
	return NIL;
}

static LISP set_point_col(LISP col)
{
	int n;

	if (NFLONUMP(col))
		err("wta(1st) to set-point-col", col);
	n = FLONM(col);
	w_list->point_pos.col = n;
	return NIL;
}

static LISP get_mark_row()
{
	return flocons(w_list->buf->mark_pos.row);
}

static LISP get_mark_col()
{
	return flocons(w_list->buf->mark_pos.col);
}

static LISP set_mark_row(LISP row)
{
	int n;
	if (NFLONUMP(row))
		err("wta(1st) to set-mark-row", row);
	n = FLONM(row);
	w_list->buf->mark_pos.row = n;
	return NIL;
}

static LISP set_mark_col(LISP col)
{
	int n;
	if (NFLONUMP(col))
		err("wta(1st) to set-mark-col", col);
	n = FLONM(col);
	w_list->buf->mark_pos.col = n;
	return NIL;
}

static LISP line_last_used_fun()
{
	int n = line_last_used(w_list->buf->matrix);
	return flocons(n);
}

static LISP col_last_used_fun(LISP row)
{
	int n;
	if (NFLONUMP(row))
		err("wta(1st) to col-last-used", row);
	n = FLONM(row);
	return flocons(col_last_used(w_list->buf->matrix, n));
}

static LISP max_lines_fun()
{
	return flocons(max_lines);
}

static LISP max_columns_fun()
{
	return flocons(max_columns);
}

/* Set up the table of functions and names */

/* Commands that take no arguments */
_fn_table fn_table[] =
{
	{"listsel-test", listsel_test},

	/* moving around */

	{"scroll-up", scroll_up},
	{"scroll-cell-up", scroll_cell_up},
	{"scroll-down", scroll_down},
	{"scroll-cell-down", scroll_cell_down},
	{"scroll-left", scroll_left},
	{"scroll-cell-left", scroll_cell_left},
	{"scroll-right", scroll_right},
	{"scroll-cell-right", scroll_cell_right},
	{"what-cursor-position", what_cursor_position},
	{"go-to", go_to},
	{"set-cell-width", set_cell_width},
	{"set-cell-height", set_cell_height},
	{"set-cell-format", set_cell_format},
	{"set-block-format", set_block_format},

	/* editing */
	{"delete-cell", delete_cell},
	{"insert-line", insert_line},
	{"remove-line", remove_line},
	{"insert-col", insert_col},
	{"remove-col", remove_col},
	{"edit-label", edit_label},
	{"edit-expression", edit_expression},

	/* block commands */
	{"copy-block", copy_block},
	{"intelligent-copy-block", intelligent_copy_block},
	{"delete-block", delete_block},
	{"preview-block", preview_block},
	{"print-block", print_block},
	{"set-block", set_block},
	{"save-block-file", save_block_file},

	/* new window */
	{"delete-window", delete_window},
	{"delete-other-windows", delete_other_windows},
	{"split-window-vertically", split_window_vertically},
	{"other-window", other_window},
	{"new-siag", new_siag},

	/* buffers and windows */
	{"switch-to-buffer", switch_to_buffer},
	{"kill-buffer", kill_buffer},
	{"load-buffer", load_buffer},
	{"save-buffer", save_buffer},
	{"save-buffer-as", save_buffer_as},

	/* help commands */
	{"print-version", print_version},
	{"help-contents", help_contents},
	{"help-copyright", help_copyright},
	{"help-search", no_op},
	{"help-for-help", help_for_help},

	/* screen layout */
	{"recenter", recenter},

	/* misc. */
	{"recalc-matrix", recalc_matrix},

	{"execute-extended-command", execute_extended_command},
	{"keyboard-quit", keyboard_quit},
	{"quit-siag", quit_siag},
	{"no-op", no_op},

	/* low level functions */
	{"get-point-row", get_point_row},
	{"get-point-col", get_point_col},
	{"get-mark-row", get_mark_row},
	{"get-mark-col", get_mark_col},
	{"line-last-used", line_last_used_fun},
	{"max-lines", max_lines_fun},
	{"max-columns", max_columns_fun},
	{(char *) 0, (LISP(*)())0}
};

/* Commands that take 1 argument */

_fn_table fn_table1[] = {
	{"auto-recalc", auto_recalc},
	{"block-borders", block_borders},
	{"plot-block", plot_block},

	/* low level functions */
	{"set-point-row", set_point_row},
	{"set-point-col", set_point_col},
	{"set-mark-row", set_mark_row},
	{"set-mark-col", set_mark_col},
	{"col-last-used", col_last_used_fun},
	{NULL, NULL}
};

void init_cmds()
{
	int i;

	for (i = 0; fn_table[i].name; i++)
		init_subr_0(fn_table[i].name, fn_table[i].function);
	for (i = 0; fn_table1[i].name; i++)
		init_subr_1(fn_table1[i].name, fn_table1[i].function);
}

