/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@asogy.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * railway.c
 */

#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/wait.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>

#include "types.h"
#include "opcodes.h"

#include "calc.h"
#include "../siod/siod.h"

/* C expression evaluator */
#define READCHAR (c=*calcexpr++)
#define READ_RETURN(t) return(READCHAR,(t))

#define val(x) ((int)((x) - '0'))
#define empty (opind == 0)
#define push(e) (opstack[++opind] = (e))
#define pop (opstack[opind--])
#define top (opstack[opind])
#define pushresstack(x) (resstack[++resind] = (x))

#define isoperator prio

static int prio[END + 1];
static int isinfix[END + 1];

static int opind, resind;
static int opstack[40];
static double resstack[40];

static double symbol_val;

char *calcexpr;			/* the expression being calculated */
int calcrow, calccol;		/* the cell being calculated */
static int c;			/* the last character read */

int errorflag;

static void init_C_parser()
{
	static int p = 0;

	prio[LPAR] = p++;
	prio[SEMICOLON] = prio[COLON] = prio[COMMA] = p++;
	prio[LOR] = p++;
	prio[LAND] = p++;
	prio[BOR] = p++;
	prio[XOR] = p++;
	prio[BAND] = p++;
	prio[EQUAL] = prio[NE] = p++;
	prio[LT] = prio[LE] = prio[GT] = prio[GE] = p++;
	prio[RSHIFT] = prio[LSHIFT] = p++;
	prio[ADD] = prio[SUB] = p++;
	prio[MULT] = prio[DIV] = prio[MOD] = p++;
	prio[LNOT] = prio[BNOT] = prio[PLUS] = prio[MINUS] = p++;

	isinfix[LPAR] = isinfix[SEMICOLON] = isinfix[COLON] = isinfix[COMMA] =
		isinfix[LOR] = isinfix[LAND] = isinfix[BOR] = isinfix[XOR] = isinfix[BAND] =
		isinfix[EQUAL] = isinfix[NE] =
		isinfix[LT] = isinfix[LE] = isinfix[GT] = isinfix[GE] =
		isinfix[RSHIFT] = isinfix[LSHIFT] =
		isinfix[ADD] = isinfix[SUB] = isinfix[MULT] = isinfix[DIV] = isinfix[MOD] =
		isinfix[LNOT] = isinfix[BNOT] = isinfix[PLUS] = isinfix[MINUS] = TRUE;
}

/* the new scanner */
static int next_symbol()
{
	char *p;
	int lineno;
	char tokentext[256];
	char *endptr;

Restart:	/* restart efter comment */
	while (isspace(c)) {
		if (c == '\n')
			lineno++;
		READCHAR;
	}
	switch (c) {
	case ',':
		READ_RETURN(COMMA);
	case '=':
		if (READCHAR)
			READ_RETURN(EQUAL);
		return ASSIGN;
	case '+':
		switch (READCHAR) {
		case '=':
			READ_RETURN(ADD_ASS);
		case '+':
			READ_RETURN(INC);
		default:
			return ADD;
		}
	case '-':
		switch (READCHAR) {
		case '=':
			READ_RETURN(SUB_ASS);
		case '-':
			READ_RETURN(DEC);
		case '>':
			READ_RETURN(PTRTOMB);
		default:
			return SUB;
		}
	case '*':
		if (READCHAR == '=')
			READ_RETURN(MULT_ASS);
		return MULT;
	case '/':
		switch (READCHAR) {
		case '=':
			READ_RETURN(DIV_ASS);
		case '*':
			READCHAR;
			while (c != '*' || READCHAR != '/') {
				if (c == '\0') {
					errorflag = TRUE;
					return NOSYMBOL;
				}
				if (c != '*')
					READCHAR;
			}
			READCHAR;
			goto Restart;
		default:
			return DIV;
		}
	case '%':
		if (READCHAR == '=')
			READ_RETURN(MOD_ASS);
		return MOD;
	case '>':
		switch (READCHAR) {
		case '>':
			if (READCHAR == '=')
				READ_RETURN(RSHIFT_ASS);
			return RSHIFT;
		case '=':
			READ_RETURN(GE);
		default:
			return GT;
		}
	case '<':
		switch (READCHAR) {
		case '<':
			if (READCHAR == '=')
				READ_RETURN(LSHIFT_ASS);
			return LSHIFT;
		case '=':
			READ_RETURN(LE);
		default:
			return LT;
		}
	case '&':
		switch (READCHAR) {
		case '=':
			READ_RETURN(BAND_ASS);
		case '&':
			READ_RETURN(LAND);
		default:
			return BAND;
		}
	case '|':
		switch (READCHAR) {
		case '=':
			READ_RETURN(BOR_ASS);
		case '|':
			READ_RETURN(LOR);
		default:
			return BOR;
		}
	case '^':
		if (READCHAR == '=')
			READ_RETURN(XOR_ASS);
		return XOR;
	case '?':
		READ_RETURN(QMARK);
	case ':':
		READ_RETURN(COLON);
	case '!':
		if (READCHAR == '=')
			READ_RETURN(NE);
		return LNOT;
	case '~':
		READ_RETURN(BNOT);
	case '(':
		READ_RETURN(LPAR);
	case ')':
		READ_RETURN(RPAR);
	case '[':
		READ_RETURN(LBRACK);
	case ']':
		READ_RETURN(RBRACK);
	case '{':
		READ_RETURN(LBRACE);
	case '}':
		READ_RETURN(RBRACE);
	case ';':
		READ_RETURN(SEMICOLON);
	case '\0':
		return END;
	case '0':
		calcexpr--;
		symbol_val = strtol(calcexpr, &endptr, 0);
		if (calcexpr == endptr)
			return NOSYMBOL;
		calcexpr = endptr;
		READCHAR;
		return NUMBER;
	case '.':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		/* read decimal number */
		calcexpr--;
		symbol_val = strtod(calcexpr, &endptr);
		if (calcexpr == endptr)	/* no conversion performed */
			return NOSYMBOL;
		calcexpr = endptr;
		READCHAR;
		return NUMBER;
	default:		/* read identifier or error */
		for (p = tokentext; isalnum(c) || (c == '_'); READCHAR)
			*p++ = c;
		*p = '\0';
		if (!strcmp(tokentext, "R"))
			return ROW;
		if (!strcmp(tokentext, "r"))
			return ROW;
		if (!strcmp(tokentext, "C"))
			return COLUMN;
		if (!strcmp(tokentext, "c"))
			return COLUMN;

		return IDENT;
	}
	return NOSYMBOL;
}

static void 
calculate(spread **matrix, int op)
{
	register int i;
	register double sum;
	register double tmp;
	long n1, n2;

	switch (op) {
	case SEMICOLON:
		sum = 0;
		tmp = resstack[resind--];
		for (i = resstack[resind]; i <= tmp; i++)
			sum += ret_val(matrix, i, calccol);
		resstack[resind] = sum;
		break;
	case COLON:
		sum = 0;
		tmp = resstack[resind--];
		for (i = resstack[resind]; i <= tmp; i++)
			sum += ret_val(matrix, calcrow, i);
		resstack[resind] = sum;
		break;
	case COMMA:
		tmp = resstack[resind--];
		resstack[resind] = ret_val(matrix, (int) resstack[resind], (int) tmp);
		break;
	case LOR:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 || n1;
		break;
	case LAND:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 && n1;
		break;
	case BOR:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 | n1;
		break;
	case XOR:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 ^ n1;
		break;
	case BAND:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 & n1;
		break;
	case EQUAL:
		tmp = resstack[resind--];
		resstack[resind] = resstack[resind] == tmp;
		break;
	case NE:
		tmp = resstack[resind--];
		resstack[resind] = resstack[resind] != tmp;
		break;
	case LT:
		tmp = resstack[resind--];
		resstack[resind] = resstack[resind] < tmp;
		break;
	case LE:
		tmp = resstack[resind--];
		resstack[resind] = resstack[resind] <= tmp;
		break;
	case GT:
		tmp = resstack[resind--];
		resstack[resind] = resstack[resind] > tmp;
		break;
	case GE:
		tmp = resstack[resind--];
		resstack[resind] = resstack[resind] >= tmp;
		break;
	case RSHIFT:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 >> n1;
		break;
	case LSHIFT:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 << n1;
		break;
	case ADD:
		tmp = resstack[resind--];
		resstack[resind] += tmp;
		break;
	case SUB:
		tmp = resstack[resind--];
		resstack[resind] -= tmp;
		break;
	case MULT:
		tmp = resstack[resind--];
		resstack[resind] *= tmp;
		break;
	case DIV:
		tmp = resstack[resind--];
		resstack[resind] /= tmp;
		break;
	case MOD:
		n1 = resstack[resind--];
		n2 = resstack[resind];
		resstack[resind] = n2 % n1;
		break;
	case LNOT:
		resstack[resind] = !resstack[resind];
		break;
	case BNOT:
		n1 = resstack[resind];
		resstack[resind] = ~n1;
		break;
	case PLUS:
		break;
	case MINUS:
		resstack[resind] = -resstack[resind];
		break;
	default:
		errorflag = TRUE;
	}
	if (resind < 1)
		errorflag = TRUE;
}				/* calculate */

static int operand(spread **matrix)
/* Non-zero return value means something went wrong */
{
	register int op;

      Operand:
	if (errorflag)
		return TRUE;
	switch (next_symbol()) {
	case NUMBER:
		pushresstack(symbol_val);
		break;
	case ROW:
		pushresstack((double) calcrow);
		break;
	case COLUMN:
		pushresstack((double) calccol);
		break;
	case LPAR:
		push(LPAR);
		goto Operand;
	case LNOT:
		push(LNOT);
		goto Operand;
	case BNOT:
		push(BNOT);
		goto Operand;
	case SUB:
		push(MINUS);
		goto Operand;
	case ADD:
		push(PLUS);
		goto Operand;
	default:
		return errorflag = TRUE;
	}

      Operator:
	if (errorflag)
		return TRUE;
	op = next_symbol();
      Restart:
	if (errorflag)
		return TRUE;
	if (isinfix[op]) {
		if (empty || prio[op] > prio[top]) {
			push(op);
			goto Operand;
		} else {
			calculate(matrix, pop);
			goto Restart;
		}
	} else if (op == END) {
		while (!empty) {
			if (isoperator[top])
				calculate(matrix, pop);
			else {
				return errorflag = TRUE;
			}
		}
		return FALSE;
	} else if (op == RPAR) {
		if (empty) {
			return errorflag = TRUE;
		}
		while (top != LPAR) {
			calculate(matrix, pop);
			if (empty) {
				return errorflag = TRUE;
			}
		}
		(void)pop;
		goto Operator;
	}
	return errorflag = TRUE;
}

/* End of C expression evaluator */

/* from slib.c */
extern char *repl_c_string_arg;
extern long repl_c_string_flag;
extern LISP repl_c_string_read();

static int siag_row, siag_col;
static double siag_result;
static spread **siag_matrix;

static LISP get_row()
{
	double drow = siag_row;
	return flocons(drow);
}

static LISP get_col()
{
	double dcol = siag_col;
	return flocons(dcol);
}

static LISP get_cell(LISP row, LISP col)
{
	int r, c;

	if (NFLONUMP(row))
		err("wta(1st) to get-cell", row);
	if (NFLONUMP(col))
		err("wta(2nd) to get-cell", col);
	r = (0.5 + FLONM(row));
	c = (0.5 + FLONM(col));
	return flocons(ret_val(siag_matrix, r, c));
}

static LISP siag_sum(LISP start, LISP end)
{
	double sum;
	int r, c, startr, startc, endr, endc;
	if (NFLONUMP(CAR(start)))
		err("wta(1st) to sum", CAR(start));
	if (NFLONUMP(CDR(start)))
		err("wta(1st) to sum", CDR(start));
	if (NFLONUMP(CAR(end)))
		err("wta(2nd) to sum", CAR(end));
	if (NFLONUMP(CDR(end)))
		err("wta(2nd) to sum", CDR(end));
	startr = (0.5 + FLONM(CAR(start)));
	startc = (0.5 + FLONM(CDR(start)));
	endr = (0.5 + FLONM(CAR(end)));
	endc = (0.5 + FLONM(CDR(end)));
	if (startr > endr) {
		r = startr;
		startr = endr;
		endr = r;
	}
	if (startc > endc) {
		c = startc;
		startc = endc;
		endc = c;
	}
	sum = 0;
	for (r = startr; r <= endr; r++)
		for (c = startc; c <= endc; c++)
			sum += ret_val(siag_matrix, r, c);
	return flocons(sum);
}

void init_parser(int argc, char **argv)
{
	print_welcome();
	process_cla(argc, argv, 1);
	print_hs_1();
	init_storage();
	init_subrs();
	init_trace();
	init_subr_0("row", get_row);
	init_subr_0("col", get_col);
	init_subr_2("get-cell", get_cell);
	init_subr_2("sum", siag_sum);

	/* and initialize the C parser as well */
	init_C_parser();
}


/* The best thing to do with this function is probably to modify it to
   always return False. Eventually it should be removed, along
   with every place it is invoked. */
int is_constant(char *p)
{
	return 0;
}


static void siag_puts(char *p)
{
}

static void siag_print(LISP p)
{
	if (FLONUMP(p)) {
		siag_result = FLONM(p);
	} else {
		siag_result = 0, errorflag = 1;
	}
}

double parse_C_expr(spread **matrix, char *expr, int row, int col)
{
	calcexpr = expr;
	READCHAR;
	calcrow = row;
	calccol = col;

	opind = resind = 0;
	/* index 0 will never contain anything, it's used as a safety margin */
	operand(matrix);
	return resstack[1];
}

/* Expressions starting with a '=' are evaluated as C */
double parse_expr(spread ** matrix, char *expr, int row, int col)
{
	/* mostly stolen from repl_c_string in slib.c */
	struct repl_hooks h;
	long retval;
	int want_sigint = 0;
	static int want_init = 1;

	if (*expr == '=') return parse_C_expr(matrix, expr+1, row, col);

	h.repl_puts = siag_puts;
	h.repl_read = repl_c_string_read;
	h.repl_eval = NULL;
	h.repl_print = siag_print;
	repl_c_string_arg = expr;	/* let's hope repl_driver */
	repl_c_string_flag = 0;	/* is non-destructive */
	siag_row = row;
	siag_col = col;
	siag_matrix = matrix;
	errorflag = 0;
	retval = repl_driver(want_sigint, want_init, &h);
	want_init = 0;		/* only once... */
	return siag_result;
}

/*
   This function was originally in cmds.c, but has been
   moved here because of its similarity to parse_expr.
 */
void execute(char *s)
{
	/* mostly stolen from repl_c_string in slib.c */
	struct repl_hooks h;
	long retval;
	int want_sigint = 0;
	static int want_init = 0;

	if (ok2print)
		hide_cur(w_list);

	h.repl_puts = siag_puts;
	h.repl_read = repl_c_string_read;
	h.repl_eval = NULL;
	h.repl_print = siag_print;
	repl_c_string_arg = s;	/* let's hope repl_driver */
	repl_c_string_flag = 0;	/* is non-destructive */
	siag_row = w_list->point_pos.row;
	siag_col = w_list->point_pos.col;
	siag_matrix = w_list->buf->matrix;
	if (strcmp(s, "(no-op)"))
		retval = repl_driver(want_sigint, want_init, &h);
	want_init = 0;		/* only once... */

	if (ok2print)
		show_cur(w_list);
}

/*
   int calc_matrix(spread **matrix)
   Recalculate all the cells in matrix.
   Returns 0 if successful, otherwise 1.
*/
/* Remarkably enough, this function requires no conversion whatsoever
   to go from long to double. */
/* 960727: reap zombies here */
int calc_matrix(spread ** matrix)
{
	register int row, col;
	int lr, lc;

	waitpid(-1, NULL, WNOHANG);

	errorflag = FALSE;
	lr = line_last_used(matrix);
	for (row = 1; row <= lr; row++) {
		lc = col_last_used(matrix, row);
		for (col = 1; col <= lc; col++) {
			switch (ret_type(matrix, row, col)) {
			case EXPRESSION:
				ins_data(matrix, ret_text(matrix, row, col),
					 parse_expr(matrix, ret_taddr(matrix, row, col), row, col),
					 EXPRESSION, row, col);
				break;
			case UNCONSTANT:
				/* uninitialized constant */
				ins_data(matrix, ret_text(matrix, row, col),
					 parse_expr(matrix, ret_taddr(matrix, row, col), row, col),
					 CONSTANT, row, col);
				break;
			default:
				;
			}
			if (errorflag)
				return 1;
		}
	}
	return 0;
}				/* calc_matrix */
